# Flow

Here is context about the flow/thread you are participating in.

- Name: {{ flow.name }}
{% if flow.description %}
- Description: {{ flow.description }}
{% endif %}
{% if flow.context %}
Context:
{% for key, value in flow.context.items() %}
- {{ key }}: {{ value }}
{% endfor %}
{% endif %}