# LLM Instructions

These instructions are specific to your LLM model. They must be followed to ensure compliance with the orchestrator and
other agents.

{% for instruction in instructions %}
- {{ instruction }}

{% endfor %}